/* 
   Matrix multiplication example

   Pthread domain decomposition solution

   Jim Teresco, CS 338, Williams College
   Mon Feb 17 20:25:15 EST 2003

   $Id: matmult_decomp.c,v 1.2 2003/02/19 20:31:11 terescoj Exp $
*/

/* header files needed for printf, gettimeofday, struct timeval */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <pthread.h>

/* header file for our own timer.c function diffgettime */
#include "timer.h"

/* we will multiply square matrices, how big? */
#define SIZE 750

/* our matrices - global because of limited stack space in Solaris */
double a[SIZE][SIZE], b[SIZE][SIZE], c[SIZE][SIZE]; 

/* function to compute the result of row row in c */
void do_row(int row) {
  int col, k;

  for (col=0; col<SIZE; col++) {
    
    /* initialize entry */
    c[row][col] = 0;
    
    /* perform dot product */
    for(k=0; k<SIZE; k++) {
      c[row][col] = c[row][col] + a[row][k]*b[k][col];
    }
  }
}

/* a struct that lets us pass in two ints to the thread function */
struct rowrange {
  int start;
  int finish;
};

void *worker(void *arg) {
  struct rowrange *range = (struct rowrange *)arg;
  int row;

  /* get our range from the argument */
  for (row=range->start; row<=range->finish; row++) {
    do_row(row);
  }
}

/* it's a simple program for now, we'll just put everything in main */
int main(int argc, char *argv[]) {
  double sum;

  /* counters */
  int i, j;

  /* to pass to gettimeofday to get wall clock times */
  struct timeval start, stop;

  int rc;

  /* pthread stuff */
  int numworkers;
  pthread_t *thread_ids;
  pthread_attr_t attr;

  /* for computation of row ranges for each worker */
  int rowsremaining, myrows;

  /* to allocate memory for argument to thread function */
  struct rowrange **rowranges;
  
  /* initialize matrices, just fill with junk and do this in serial */
  gettimeofday(&start, NULL);
  for (i=0; i<SIZE; i++) {
    for (j=0; j<SIZE; j++) {
      a[i][j] = i+j;
      b[i][j] = i-j;
    }
  }
  gettimeofday(&stop, NULL);
  printf("Initialization (serial) took: %f seconds\n", diffgettime(start,stop));
  
  /* number of workers is specified on the command-line */
  numworkers=1;
  if (argc > 1) numworkers=atoi(argv[1]);

  thread_ids = (pthread_t *)malloc(numworkers*sizeof(pthread_t));
  rowranges = (struct rowrange **)malloc(numworkers*sizeof(struct rowrange *));

  /* allocate and populate rowranges structures to pass to thread functions */
  rowsremaining=SIZE;
  for (i=0; i<numworkers; i++) {
    rowranges[i] = (struct rowrange *)malloc(sizeof(struct rowrange));
    myrows = rowsremaining/(numworkers-i);
    rowranges[i]->start = (SIZE-rowsremaining);
    rowranges[i]->finish = rowranges[i]->start + myrows - 1;
    rowsremaining -= myrows;
    printf("Worker %d will compute %d rows: %d-%d\n", i, myrows,
	   rowranges[i]->start, rowranges[i]->finish);
  }

  /* some pthread setup */
  pthread_setconcurrency(numworkers+1);
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

  gettimeofday(&start, NULL);
  /* start workers */
  for (i=0; i<numworkers; i++) {
    rc=pthread_create(&(thread_ids[i]), &attr, &worker, 
		      (void *)rowranges[i]);
    if (rc != 0) {
      fprintf(stderr, "Could not create thread %d\n", i);
      exit(1);
    }
  }

  /* workers working.... */

  /* wait for workers to finish */
  for (i=0; i<numworkers; i++) {
    pthread_join(thread_ids[i], NULL);
  }

  gettimeofday(&stop, NULL);
  printf("Multiplication took: %f seconds\n", diffgettime(start,stop));

  /* This is here to make sure the optimizing compiler doesn't
     get any big ideas about "optimizing" code away completely */
  sum=0;
  for (i=0; i<SIZE; i++) {
    for (j=0; j<SIZE; j++) {
      sum += c[i][j];
    }
  }
  printf("Sum of elements of c=%f\n", sum);

  free(thread_ids);
  for (i=0; i<numworkers; i++) {
    free(rowranges[i]);
  }
  free(rowranges);

  return 0;
}
